"""
    Created on 2012-01-20
    @author: jldupont
"""
import os, sys, logging, re
from time import sleep

from tools_logging import pprint_kv
from tools_s3      import json_string
from tools_os      import resolve_path, mkdir_p, gen_walk, remove_common_prefix, safe_path_exists, touch
from tools_os      import can_write, rm
from tools_sys     import retry, move

import boto
from boto.s3.key import Key as S3Key

def stdout(s):
    sys.stdout.write(s+"\n")

def report(ctx, ctx2):
    d=dict(ctx2)
    d.update(ctx)
    stdout(json_string(d)) 


def run(enable_simulate=False, bucket_name=None, 
        path_source=None, path_moveto=None, path_check=None,
        num_files=5, enable_delete=False, propagate_error=False, prefix=None, polling_interval=None
        ,only_ext=None
        ,filename_input_regex=None
        ,key_output_format=None
        ,enable_progress_report=False
        ,write_done=False
        ,**_):
    
    if key_output_format is not None:
        if filename_input_regex is None:
            raise Exception("-ifnr and -okf options work in tandem")
    
    if filename_input_regex is not None:
        logging.info("Compiling input filename regex...")
        try:
            ireg=re.compile(filename_input_regex.strip("'"))
            ofmt=key_output_format.strip("'")
        except:
            raise Exception("Can't compile input filename regex pattern")
        
        if key_output_format is None:
            raise Exception("Input filename regex specified but no output S3 key format specified")
    else:
        ireg=None
        ofmt=None
    
    
    #if args.enable_debug:
    #    logger=logging.getLogger()
    #    logger.setLevel(logging.DEBUG)
    
    bucket_name=bucket_name.strip()
    path_source=path_source.strip()
    
    try:    prefix=prefix.strip()
    except: prefix=None
    
    try:    path_moveto=path_moveto.strip()
    except: path_moveto=None
    
    if path_check is not None:
        code, path_check=resolve_path(path_check)
        if not code.startswith("ok"):
            logging.warning("path_check '%s' might be in error..." % path_check)
    
    ### VALIDATE PARAMETERS
    if enable_delete and path_moveto is not None:
        raise Exception("-d can't be used with -m")
    
    code, p_src=resolve_path(path_source)
    if not code.startswith("ok"):
        raise Exception("Invalid source path: %s" % path_source)
    
    if path_moveto is not None:
        code, p_dst=resolve_path(path_moveto)
        if not code.startswith("ok"):
            raise Exception("Invalid moveto path: %s" % path_moveto)
    else:
        p_dst=None
    
    ### wait for 'source' path to be available
    logging.info("Waiting for source path to be accessible... CTRL-c to stop")
    while True:
        if os.path.isdir(p_src):
            break
        sleep(1)
    logging.info("* Source path accessible")

    if path_moveto is not None:
        logging.info("Creating 'moveto' directory if required")
        code, _=mkdir_p(p_dst)
        if not code.startswith("ok"):
            raise Exception("Can't create 'moveto' directory: %s" % p_dst)
        logging.info("* Created moveto directory")
    
    if not enable_simulate:
        try:
            conn = boto.connect_s3()
        except:
            ## not much we can do
            ## but actually no remote calls are made
            ## at this point so it should be highly improbable
            raise Exception("Can't 'connect' to S3")
    
    if not enable_simulate:
        ###################### BUCKET
        logging.info("Getting/creating bucket (unlimited retries with backoff)")
        def _get_create_bucket():
            return conn.create_bucket(bucket_name)
                  
        bucket=retry(_get_create_bucket)
        logging.info("Got bucket")
        #############################

    if enable_simulate:
        logging.info("Begin simulation...")
    else:
        logging.debug("Starting loop...")

    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid:  %s" % ppid)
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
        #################################################

        _code, path_exists=safe_path_exists(path_check)
                        
        if path_check is None or path_exists:
            try: 
                logging.debug("Start processing...")
                count=0
                gen=gen_walk(p_src, max_files=num_files,only_ext=only_ext)
                    
                for src_filename in gen:

                    if enable_progress_report:
                        logging.info("Processing file: %s" % src_filename)
                    
                    if write_done:
                        if is_done_file(src_filename):
                            continue
                    
                    try:          
                        s3key_name=gen_s3_key(ireg, ofmt, p_src, src_filename, prefix)
                    except Exception,e:
                        raise Exception("Error generating S3 key... check your command line parameters... use the 'simulate' facility: %s" % e)
                    
                    if enable_simulate:
                        simulate(src_filename, s3key_name, enable_delete, p_dst)
                    else:
                        k=S3Key(bucket)
                        k.key=s3key_name
                        was_uploaded=process_file(enable_progress_report, bucket_name, prefix, k, src_filename, p_dst, enable_delete, propagate_error)
                        if was_uploaded:
                            count=count+1
                            if write_done:
                                do_write_done(src_filename)
    
            except Exception, e:
                logging.error("Error processing files...(%s)" % str(e))
        else:
            logging.info()

        if count>0:
            logging.info("Progress> uploaded %s files" % count)

        #####################################################
        logging.debug("...sleeping for %s seconds" % polling_interval)
        sleep(polling_interval)

    
def is_done_file(filename):
    return filename.endswith("done")

    
def do_write_done(src_filename):
    dfile="%s.done" % src_filename
    code, _=touch(dfile)
    if code!="ok":
        raise Exception("Can't write '.done' file: %s" % dfile)

def gen_s3_key(ireg, ofmt, p_src, filename, prefix):
    _, fn=remove_common_prefix(p_src, filename)
    
    ### use the input regex and output format string
    ### to generate S3 key
    if ireg is not None:
        try:
            g=ireg.match(fn).groups()
            fn=ofmt % g
        except:
            raise Exception("Not match found using input regex, filename: %s" % filename)
    
    s="/%s%s" % (prefix, fn)
    return s.replace("//","")
    
    
    
def simulate(fil, s3key_name, enable_delete, p_dst):
    pprint_kv("File to be uploaded", fil)
    pprint_kv(" Filename used on s3", s3key_name)
    
    if enable_delete:
        _c, f_is_writable=can_write(fil)      
        pprint_kv(" File would be deleted", fil)
        if not f_is_writable:
            pprint_kv(" ! File can't be deleted", fil)
    else:
        if p_dst is not None:
            bname=os.path.basename(fil)
            fdst=os.path.join(p_dst, bname)        
            _c, d_is_writable=can_write(p_dst)
            pprint_kv(" File would be moved to", fdst)
            if not d_is_writable:
                pprint_kv(" ! File can't be moved to", p_dst)


def process_file(enable_progress_report, bucket_name, prefix, k, src_filename, p_dst, enable_delete, propagate_error):
    
    uploaded=False
    ctx={"src": src_filename, "key": k.name, "bucket": bucket_name, "prefix": prefix}
    
    #1) Upload to S3
    try:
        k.set_contents_from_filename(src_filename)
        report(ctx, {"code":"ok", "kind":"upload"})
        
        if enable_progress_report:
            logging.info("progress: uploaded file %s" % src_filename)
            uploaded=True
    except:
        if propagate_error:
            report(ctx, {"code":"error", "kind":"upload"})
        return
    
    #2a) Delete
    if enable_delete:
        code, msg=rm(src_filename)
        if not code.startswith("ok"):
            logging.debug("Error deleting: %s (%s)" % (src_filename, msg))
            if not propagate_error:
                return
        else:
            if enable_progress_report:
                logging.info("progress: deleted file %s" % src_filename)

        report(ctx, {"code": code, "kind":"delete"})
            
    #2b) Move
    else:
        bname=os.path.basename(src_filename)
        dst_filename=os.path.join(p_dst, bname)
        code, msg=move(src_filename, dst_filename)
        if not code.startswith("ok"):
            ## 1 last chance... try recreating the dst directory for next run...
            mkdir_p(p_dst)
            logging.debug("Error moving: %s ==> %s  (%s)" % (src_filename, dst_filename, msg))
            if not propagate_error:
                return
        else:
            if enable_progress_report:
                logging.info("progress: moved file %s ==> %s" % (src_filename, dst_filename))

        report(ctx, {"code":code, "kind":"move", "dst":dst_filename})
    
    return uploaded




"""
    Created on 2012-01-20
    @author: jldupont
"""
import logging, os, json
from time import sleep

from pyfnc.fnc_tools import coroutine

from jldaws.tools_os import mkdir_p, resolve_path, get_root_files, file_contents, rm
from jldaws.tools_sys import retry
from jldaws.db.skv import SimpleKV

def run(domain_name=None
        ,src_path=None
        ,polling=None 
        ,batch_size=None
        ,category_name=None
        ,default_value=None
        ,src_ext=None
        ,reset_count=None
        ,**_
        ):

    code, src_path=resolve_path(src_path)
    if not code.startswith("ok"):
        raise Exception("Can't resolve source path...")

    logging.info("Creating source path if required...")
    mkdir_p(src_path)

    ######################
    
    db=SimpleKV()
    def connect():
        db.connect()
    
    logging.info("Connecting to SDB... CTRL-C to abort")
    retry(connect)
    
    def create():
        db.set_domain(domain_name)
        
    logging.info("Creating domain: %s... CTRL-C to abort" % domain_name)
    retry(create)
    
    #############
    stage1=process1(db, batch_size, category_name, default_value)
    pollcount=0
    
    # MAIN LOOP
    ###########
    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid:  %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        reset=(pollcount==reset_count)
            
        code, files=get_root_files(src_path, only_ext=src_ext)
        if code.startswith("ok"):
            logging.debug("Got files: %s" % files)
            stage1.send((reset, files))
            
        logging.debug("Sleeping...")
        sleep(polling)
        pollcount=pollcount+1
        

@coroutine
def process1(db, batch_size, category_name, default_value):

    stage2=process2(db, category_name, default_value)
    recent=[]
    
    while True:
        reset, files=(yield)
        
        #qlogging.debug("stage1: received: %s %s" % (reset, files))
        
        if reset:
            recent=[]

        def filter_recent(_file):
            return _file not in recent

        files=filter(filter_recent, files)
            
        files=files[0:batch_size]
        if len(files)!=0:            
            logging.info("Progress> processing %s files" % len(files))
        
            for _file in files:
                recent.append(_file)
                
                code, maybe_contents=file_contents(_file)
                if code.startswith("ok"):
                    stage2.send((_file, maybe_contents))
                else:
                    logging.warning("Can't read contents of: %s" % _file)
            

@coroutine
def process2(db, category_name, default_value):
    
    stage3=process3()
    
    while True:
        _file, contents=(yield)
        logging.info("Progress> processing file: %s" % _file)
    
        inserted=False    
        try:
            ino=json.loads(contents)
            ino["category"]=ino.get("category", category_name)
            ino["value"]=ino.get("value", default_value)
            if ino.get("key", None) is None:
                logging.warning("Missing 'key' field in file: %s" % _file)
            else:
                logging.debug("Input JSON: %s" % ino)
                inserted=do_insert(db, ino["category"], ino["key"], ino["value"])
        except:
            logging.warning("Error processing file: %s" % _file)

        if inserted:
            stage3.send(_file)
        
    
                
def do_insert(db, category, key, value):
    logging.debug("Inserting [%s:%s:%s]" % (category, key, value))
    try:
        db.insert(key, value, category=category)
        logging.info("Progress: inserted 1 record")
        return True
    except:
        try:
            ### don't give up too easily
            sleep(1)
            db.insert(key, value, category=category)
            logging.info("Progress: inserted 1 record (after 2nd try)")
            return True
        except Exception,e:
            logging.debug(e)
            logging.warning("Can't insert in SDB...")
            
    return False
          
@coroutine
def process3():
    """
    Delete source file once processed
    """
    while True:
        
        _file_to_delete=(yield)
        rm(_file_to_delete)
        

    



"""
    Created on 2012-01-19
    @author: jldupont
"""
import uuid

from tools_date import get_timestamp

def gen_queue_name():
    """
    >>> _gen_queue_name()
    """
    _, ts=get_timestamp()
    queue_name="ex_%s_%s" % (ts, uuid.uuid1())
    return queue_name

if __name__=="__main__":
    import doctest
    doctest.testmod()



"""
    @author: Jean-Lou Dupont
"""

import sys,os

## force stdout + stderr flushing
sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 0)

from tools_logging import setup_basic_logging

setup_basic_logging()




"""
    Created on 2012-01-19
    @author: jldupont
"""
import os, errno,re, tempfile
import subprocess


def atomic_write(path, contents, tmppath=None):
    """
    Atomic write to file
    
    Create temporary file and then move/rename to specified path.
    Rename operation in the same filesystem are atomic (at least in Linux).
    
    >>> atomic_write("/tmp/_jlddk_atomic_write", "test!") ## doctest: +ELLIPSIS
    ('ok', ...
    """
    
    dn=tmppath or os.path.dirname(path)
    fd, tfn=tempfile.mkstemp(dir=dn)
    
    try:
        ### part 1: write to temp file
        f=os.fdopen(fd, "w")
        f.write(contents)
        f.close()
    except Exception, e:
        try:    os.close(fd)
        except: pass
        return ("error", "write to temp file: %s" % str(e))        
        
    try:
        ### part 2: rename to specified path
        os.rename(tfn, path)
    except Exception,e:
        return ("error", "rename to path '%s': %s" % (path, e))
        
    return ("ok", tfn)


def filter_files(target_ext, files):
    if target_ext is None:
        return files
        
    def inf(path):
        _root, ext=os.path.splitext(path)
        return ext.strip(".")==target_ext
    
    return filter(inf, files)


def get_root_files(src_path, strip_dirname=False, only_ext=None):
    """
    Retrieve files from the root of src_path
    >> get_root_files("/tmp")
    
    >> get_root_files("/tmp", strip_dirname=True)
    
    """
    def add_src_path(p):
        return os.path.join(src_path, p)
    
    
    try:
        liste=os.listdir(src_path)
        liste=map(add_src_path, liste)        
        liste=filter(os.path.isfile, liste)
        
        if strip_dirname:
            liste=map(os.path.basename, liste)

        liste=filter_files(only_ext, liste)
                 
        return ("ok", liste)
    except:
        return ("error", None)


RE_EXT=re.compile(r'^(.*?)\-(.*?)\.([a-zA-Z]+|[a-zA-Z]+\.[a-zA-Z]+)$')

def split_path_version(path):
    """
    >>> print split_path_version("file.ext")
    (None, None, None)
    >>> print split_path_version("file-0.1.ext")
    ('file', '0.1', 'ext')
    >>> print split_path_version("file-0.1.1.ext")
    ('file', '0.1.1', 'ext')
    >>> print split_path_version("file-0.1.1.tar.gz")
    ('file', '0.1.1', 'tar.gz')
    >>> print split_path_version("file-0.1.1a.tar.gz")
    ('file', '0.1.1a', 'tar.gz')
    >>> print split_path_version("file-0.1b.1a.tar.gz")
    ('file', '0.1b.1a', 'tar.gz')
    >>> print split_path_version("file-0.1b.1-dev.tar.gz")
    ('file', '0.1b.1-dev', 'tar.gz')
    >>> print split_path_version("file-0.4.tar.gz")
    ('file', '0.4', 'tar.gz')
    """
    try:
        groups=RE_EXT.match(path).groups()
        return groups
    except:
        return (None, None, None)
    


def safe_can_write(path):
    """
    Check if 'path' can be written to
    
    - If 'path' exists, do safe test (i.e. non-destructive)
    - If 'path' !exists, do 'append' test
    
    >>> safe_can_write("/tmp/!+$,@")
    ('ok', True)
    >>> rm("/tmp/!+$,@")
    ('ok', '/tmp/!+$,@')
    >>> safe_can_write("/tmp/")
    ('ok', True)
    """
    try:
        maybe_exists=os.path.exists(path)
        if maybe_exists:
            return can_write(path)
    except:
        pass

    try:
        f=open(path, "a")
    except:
        return ("error", False)
    finally:
        try:
            rm(path)
            f.close()
        except:
            pass
    
    return ("ok", True)

def check_can_write_paths(src_dir, paths):
    """
    >>> check_can_write_paths("/tmp", ["allo", "!$[].+$", ".X11-unix"])
    [('ok', ('allo', True)), ('ok', ('!$[].+$', True)), ('ok', ('.X11-unix', True))]
    """
    results=[]
    for path_fragment in paths:
        try:
            path=os.path.join(src_dir, path_fragment)
            code, resp=safe_can_write(path)
            
            results.append((code, (path_fragment, resp)))
        except Exception,e:
            results.append(("error", (path_fragment, e)))
            
    return results
    

def check_exist_paths(src_dir, paths):
    """
    >>> check_exist_paths("/tmp", ["allo", "!$[].+$", ".X11-unix"])
    [('ok', ('allo', False)), ('ok', ('!$[].+$', False)), ('ok', ('.X11-unix', True))]
    """
    results=[]
    for path_fragment in paths:
        try:
            path=os.path.join(src_dir, path_fragment)
            exists=os.path.exists(path)
            results.append(("ok", (path_fragment, exists)))
        except Exception,e:
            results.append(("error", (path_fragment, e)))
            
    return results
        


def safe_path_exists(path):
    try:
        return ("ok", os.path.exists(path))
    except:
        return ("error", False)

def touch(path):
    """
    >>> touch("/tmp/JustTouching")
    ('ok', '/tmp/JustTouching')
    >>> rm('/tmp/JustTouching')
    ('ok', '/tmp/JustTouching')
    """
    fhandle = file(path, 'a')
    try:
        os.utime(path, None)
        return ('ok', path)
    except OSError, exc:
        return ("error", (exc.errno, errno.errorcode[exc.errno]))
    finally:
        fhandle.close()        

def rm(path):
    """
    Silently (i.e. no exception thrown) removes a path if possible
    
    >>> rm("/tmp/NotAFile") ## no need to complain if there is no file
    ('ok', '/tmp/NotAFile')
    """
    try:
        os.remove(path)
        return ('ok', path)
    except OSError, exc:
        if exc.errno==errno.ENOENT:
            return ('ok', path)
        return ("error", (exc.errno, errno.errorcode[exc.errno]))

def can_write(path):
    """
    Checks if the current user (i.e. the script) can delete the given path
    Must check both user & group level permissions
    
    @return: (code, path | msg)
    
    FOR THIS TEST, NEED TO CREATE A DIRECTORY /tmp/_test_root_sipi THROUGH ROOT
    
    >>> p="/tmp/_test_sipi"
    >>> rm(p)
    ('ok', '/tmp/_test_sipi')
    >>> touch(p)
    ('ok', '/tmp/_test_sipi')
    >>> can_write(p)
    ('ok', True)
    >>> rm(p)
    ('ok', '/tmp/_test_sipi')
    >>> can_write("/tmp/_test_root_sipi/")
    ('ok', False)
    """
    try:
        return ("ok", os.access(path, os.W_OK))
    except:
        return ("error", None)



def remove_common_prefix(common_prefix, path):
    """
    >>> remove_common_prefix("/tmp", "/tmp/some_dir/some_file.ext")
    ('ok', '/some_dir/some_file.ext')
    """
    try:
        _head, _sep, tail=path.partition(common_prefix)
        return ("ok", tail)
    except:
        return ("error", path)


def gen_walk(path, max_files=None, only_ext=None):
    """
    >>> g=gen_walk("/tmp/test")
    >>> for f in g:
    ...    print f
    """
    count=0
    done=False
    for root, _dirs, files in os.walk(path):
        
        for f in files:
            if f is None:
                continue
            if only_ext is not None:
                if not f.endswith(only_ext):
                    continue
            yield os.path.join(root, f)
        
            count=count+1
            if max_files is not None:
                if count==max_files:
                    done=True
                    break
            
        if done: break
        



def safe_listdir(path):
    try:
        return ("ok", os.listdir(path))
    except Exception, e:
        return ("error", e)

def get_path_extension(path):
    """
    >>> get_path_extension("/some/path/config.yaml")
    ('/some/path/config', '.yaml')
    >>> get_path_extension(".bashrc")
    ('.bashrc', '')
    >>> get_path_extension("/some/path/file")
    ('/some/path/file', '')
    """
    return os.path.splitext(path)

def mkdir_p(path):
    """
    Silently (i.e. no exception thrown) makes a directory structure if possible
    
    This function can be found in the 'sipi' package.
    """
    try:
        os.makedirs(path)
        return ('ok', path)
    except OSError, exc:
        if exc.errno == errno.EEXIST:
            return ('ok', path)

        return ('error', (exc.errno, errno.errorcode[exc.errno]))


def resolve_path(path):
    try:
        path=os.path.expandvars(path)
        path=os.path.expanduser(path)
        return ("ok", path)
    except:
        return ("error", None)


def safe_isfile(path):
    try:
        return os.path.isfile(path)
    except:
        return False
    
    
def simple_popen(path_script, env={}):
    new_env=os.environ.copy()
    new_env.update(env)
    return subprocess.Popen(path_script, env=new_env)

    
    
def file_contents(path):
    """
    Simple "get file contents"
    """
    try:
        fh=open(path, "r")
        return ('ok', fh.read())
    except:
        return ("error", None)
    finally:
        try:
            fh.close()
        except:
            pass

if __name__=="__main__":
    import doctest
    doctest.testmod()
    


"""
    Created on 2012-01-20
    @author: jldupont
"""
import logging, os, sys, json

import boto
from boto.s3.key import Key as S3Key

from jldaws.tools_sys import retry

def run(bucket_name=None, bucket_prefix=None, 
        alternate_format=False,
        just_basename=False,
        trigger_topic=None,
        **_
        ):

    try:
        conn = boto.connect_s3()
    except:
        ## not much we can do
        ## but actually no remote calls are made
        ## at this point so it should be highly improbable
        raise Exception("Can't 'connect' to S3")

    logging.info("Creating bucket...")
    
    def bucket_create():
        return conn.create_bucket(bucket_name)

    try:
        bucket=retry(bucket_create)
        logging.info("Got bucket '%s'" % bucket_name)
    except:
        raise Exception("Can't get bucket '%s'" % bucket_name)
    
    
    # MAIN LOOP
    ###########
    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid: %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break

        line=sys.stdin.readline().strip()

        if trigger_topic is not None:
            if len(line)==0:
                continue
            try:
                jo=json.loads(line)
                logging.debug("Loaded JSON...")
            except:
                logging.error("Can't load JSON from line: %s" % line)
                continue
        
            try:
                topic=jo["topic"]
                logging.debug("Extracted topic: %s" % topic)
            except:
                logging.error("JSON object doesn't have 'topic' field")
                continue
        
            if topic!=trigger_topic:
                logging.debug("Topic != trigger_topic: %s" % topic)
                continue
            
        logging.debug("Triggered...")
        
        
        """
            s3://bucket_name/bucket_prefix/key
            s3://bucket_name/key
            bucket_name /t bucket_prefix/key
        """
        
        if alternate_format:
            _base_format="%s\t" % bucket_name
        else:
            _base_format="s3://%s/" % bucket_name
        
        if bucket_prefix is not None:
            liste=bucket.list(prefix=bucket_prefix)
        else:
            liste=bucket.list()
            
        
        for key in liste:
            if not key.name.endswith("/"):
                if just_basename:
                    print os.path.basename(key.name)
                else:
                    print "%s%s" % (_base_format, key.name)



"""
    Created on 2012-01-20
    @author: jldupont
"""
import logging, os, json, signal
from time import sleep
import uuid

from tools_os import can_write, rm, touch
from tools_sys import retry, SignalTerminate, jstdout
#from tools_mod import prepare_callable
from tools_sqs import gen_queue_name
from tools_sns import build_topic_arn
from tools_leader import protocol_processor

import boto
from boto.sqs.message import RawMessage
from boto.exception import SQSDecodeError

sigtermReceived=False

def handlerSigTerm(*p, **k):
    global sigtermReceived
    sigtermReceived=True

def run(dst_path=None, topic_name=None, queue_name=None, mod_sub=None, mod_pub=None, 
        polling_interval=None, force=False, node_id=None, force_delete=False, delete_queue=None,
        proto_n=None, proto_m=None
        ,**_):

    signal.signal(signal.SIGTERM, handlerSigTerm)

    ### STARTUP CHECKS
    ##################
    
    if proto_n > proto_m:
        raise Exception("Parameter 'n' must be smaller than 'm'")
    
    if os.path.isdir(dst_path):
        raise Exception("'dst_path' must be a filename, not a directory")

    if force_delete:
        logging.info("Attempting to delete '%s'" % dst_path)
        rm(dst_path)

    if os.path.isfile(dst_path):
        raise Exception("'dst_path' must not exists at startup: use -fd to delete")
        
    dir_path=os.path.dirname(dst_path)
    code, _msg=can_write(dir_path)
    if not code.startswith("ok"):
        raise Exception("directory '%s' is not writable" % dir_path)
    
    ### SETUP
    ###########################
    if node_id is None:
        node_id=str(uuid.uuid1())
        
    logging.info("Node id: %s" % node_id)

    proc=protocol_processor(node_id, proto_n, proto_m)
    
    
    ### START MAIN LOOP
    ######################################
    run_aws(node_id, proc, polling_interval, queue_name, topic_name, dst_path, delete_queue)
    """
    if force:
        run_force(node_id, proc, polling_interval, dst_path)
    else:
        if mod_pub is None or mod_sub is None:
            run_aws(node_id, proc, polling_interval, queue_name, topic_name, dst_path)
        else:
            run_mod(node_id, proc, polling_interval, topic_name, dst_path, mod_sub, mod_pub)
    """

        
def run_aws(node_id, proc, polling_interval, queue_name, topic_name, dst_path, delete_queue):
    
    if topic_name is None:
        raise Exception("Need a topic_name")
    
    auto_queue=False
    if queue_name is None:
        auto_queue=True
        queue_name=gen_queue_name()

    def setup_private_queue():
        conn = boto.connect_sqs()
        q=conn.create_queue(queue_name)
        q.set_message_class(RawMessage)
        return (conn, q) 

    # SETUP PRIVATE QUEUE
    logging.info("Creating queue '%s'" % queue_name)
    sqs_conn, q=retry(setup_private_queue, logmsg="Having trouble creating queue...")
    
    topic_arn=build_topic_arn(sqs_conn, topic_name)
    
    logging.info("topic_arn: %s" % topic_arn)    
    
    ### create topic
    def create_topic():
        """
        {'CreateTopicResponse': 
            {'ResponseMetadata': 
                {'RequestId': '5e2c6700-4dd0-11e1-b421-41716ce69b95'}, 
            'CreateTopicResult': {'TopicArn': 'arn:aws:sns:us-east-1:674707187858:election'}}}
        """
        snsconn=boto.connect_sns()
        snsconn.create_topic(topic_name)
        
    retry(create_topic, logmsg="Having trouble creating topic...")
    
    # SUBSCRIBE TO TOPIC
    def sub_topic():
        snsconn=boto.connect_sns()
        snsconn.subscribe_sqs_queue(topic_arn, q)
        return snsconn

    snsconn=retry(sub_topic, logmsg="Having trouble subscribing queue to topic...")
    
    logging.info("Subscribed to topic '%s'" % topic_name)
        
    current_state="NL"
    
    MSGS={"NL": "Leadership lost",
          "L":  "Leadership gained",
          "ML": "Leadership momentum"
          }
    
    poll_count=0

    def cleanup():
        rm(dst_path)
        if auto_queue or delete_queue:
            logging.info("... deleting queue")
            try:    sqs_conn.delete_queue(q)
            except: pass
    
    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid: %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            cleanup()
            break

        try:
            ## do a bit of garbage collection :)        
            global sigtermReceived
            if sigtermReceived:
                cleanup()
                raise SignalTerminate()
            #########################################
                
            try:
                ### BATCH PROCESS - required!!!
                while True:
                    rawmsg=q.read()
                    if rawmsg is not None:
                        jsonmsg=json.loads(rawmsg.get_body())
                        q.delete_message(rawmsg)
                        
                        ## SNS encapsulates the original message...
                        nodeid=str(jsonmsg["Message"])
                        
                        transition, current_state=proc.send((poll_count, nodeid))
                        jstdout({"state": current_state})
                        
                        if transition:
                            logging.info(MSGS[current_state])
                            if current_state=="L":
                                code, _=touch(dst_path)
                                logging.info("Created '%s': %s" % (dst_path, code))
                            else:
                                code, _=rm(dst_path)
                                logging.info("Deleted '%s': %s" % (dst_path, code))
                    else:
                        break
                    
            except SQSDecodeError:
                logging.warning("Message decoding error")
                
            except Exception,e:
                logging.error(str(e))
                continue
        
            msg=str(node_id)
        
            logging.debug("Publishing our 'node id': %s" % node_id)
            try:
                snsconn.publish(topic_arn, msg)
            except:
                try:
                    snsconn.publish(topic_arn, msg)
                except:
                    logging.warning("Can't publish to topic '%s'" % topic_name)
            
            logging.debug("... sleeping for %s seconds" % polling_interval)
            sleep(polling_interval)
            poll_count=poll_count+1
            
        except KeyboardInterrupt:
            cleanup()
            raise
        
        

"""
def run_force(node_id, proc, polling_interval, dst_path):
    pass


def run_mod(node_id, proc, polling_interval, topic_name, dst_path, mod_sub, mod_pub):
    fnc_sub=prepare_callable(mod_sub, "subscribe_message")
    fnc_pub=prepare_callable(mod_sub, "publish_message")
    
    poll_count=1
    logging.info("Starting loop...")
    while True:
       
        while True:
            try:
                code, maybe_msg=fnc_sub(topic_name)
                if not code.startswith("ok"):
                    logging.error("Error retrieving message: %s" % maybe_msg)
                    break
                else:
                    if maybe_msg is not None:
                        current_state=proc.send((poll_count, maybe_msg))
            except:
                pass

        try:
            fnc_pub(topic_name, {})
        except Exception, e:
            logging.error("Publish function: %s" % str(e))
        
        logging.debug("... sleeping for %s seconds" % polling_interval)
        sleep(polling_interval)
        poll_count=poll_count+1
"""


"""
    Created on 2012-01-19
    @author: jldupont
"""
import os

try:
    import boto
except:
    raise Exception("* package 'boto' is required - get it from Pypi\n")

try:
    import argparse
except:
    raise Exception("* package 'argparse' is necessary - get it from Pypi\n")

try:
    os.environ["AWS_ACCESS_KEY_ID"]
    os.environ["AWS_SECRET_ACCESS_KEY"]
    os.environ["AWS_ACCOUNT_ID"]
except Exception, e:
    raise Exception("* environment variable missing: %s" % str(e))



"""
    Created on 2012-01-25
    @author: jldupont
"""
from nose.tools import raises


import jldaws.db.skv as skv
from jldaws.exceptions import SDB_Access

class Test():
    
    @classmethod
    def setup_class(self):
        """setup"""
        self.db=skv.SimpleKV()
        self.db.connect()
        self.db.set_domain("test")
    
    @classmethod    
    def teardown_class(self):
        """teardown"""
        
    def test_00delete_domain(self):
        self.db.delete_domain()
        self.db.set_domain("test")
        
    def test_01get_unknown(self):
        r=self.db.get("unknownkey")
        assert r is None

    def test_10insert(self):
        eid, _, _=self.db.insert("testkey", "testvalue")
        r=self.db.get("testkey", consistent_read=True)
        assert str(r.name) == str(eid)
        
    def test_11delete(self):
        count=self.db.delete("testkey")
        assert count==1
        
    def test_50category(self):
        eid, _, _=self.db.insert("catkey", "catvalue", category="cat")
        cr=self.db.get_by_category("cat", last=True, consistent_read=True)
        assert str(cr.name) == str(eid)
        
    @raises(SDB_Access)
    def test_99raises(self):
        self.db.delete_item(None, silent=False)